@@ -19,7 +19,7 @@ use Carp;
 
 =head1 NAME
 
-Mail::DKIM::DkSignature - a DomainKeys signature header
+Mail::DKIM::DkSignature - represents a DomainKeys-Signature header
 
 =head1 CONSTRUCTORS
 
@@ -126,6 +126,25 @@ sub body_hash
 	croak "body_hash not implemented";
 }
 
+=head2 algorithm() - get or set the algorithm (a=) field
+
+The algorithm used to generate the signature.
+Defaults to "rsa-sha1", an RSA-signed SHA-1 digest.
+
+=cut
+
+sub algorithm
+{
+	my $self = shift;
+
+	if (@_)
+	{
+		$self->set_tag("a", shift);
+	}
+
+	return lc $self->get_tag("a") || 'rsa-sha1';
+}	
+
 =head2 canonicalization() - get or set the canonicalization (c=) field
 
   $signature->canonicalization("nofws");
@@ -176,7 +195,9 @@ sub domain
 
 sub expiration
 {
-	croak "expiration not implemented";
+	my $self = shift;
+	croak "cannot change expiration on " . ref($self) if @_;
+	return undef;
 }
 
 use MIME::Base64;
@@ -205,6 +226,12 @@ sub check_protocol
 	return ($protocol eq "dns");
 }
 
+sub check_version
+{
+	#DomainKeys doesn't have a v= tag
+	return 1;
+}
+
 sub get_algorithm_class
 {
 	my $self = shift;
@@ -256,7 +283,7 @@ sub identity
 
 sub method
 {
-	croak "method not implemented";
+	croak "method not implemented (use canonicalization instead)";
 }	
 
 =head2 protocol() - get or set the query methods (q=) field
@@ -272,7 +299,9 @@ sub protocol {
 	(@_) and
 		$self->set_tag("q", shift);
 
-	return $self->get_tag("q");
+	# although draft-delany-domainkeys-base-06 does mandate presence of a
+	# q=dns tag, it is quote common that q tag is missing - be merciful
+	return !defined($self->get_tag("q")) ? 'dns' : $self->get_tag("q");
 }	
 
 =head2 selector() - get or set the selector (s=) field
@@ -302,4 +331,22 @@ sub version
 	croak "version not implemented";
 }
 
+=head1 SEE ALSO
+
+L<Mail::DKIM::Signature> for DKIM-Signature headers
+
+=head1 AUTHOR
+
+Jason Long, E<lt>jlong@messiah.eduE<gt>
+
+=head1 COPYRIGHT AND LICENSE
+
+Copyright (C) 2006-2007 by Messiah College
+
+This library is free software; you can redistribute it and/or modify
+it under the same terms as Perl itself, either Perl version 5.8.6 or,
+at your option, any later version of Perl 5 you may have available.
+
+=cut
+
 1;
